package io.tswf.transmutations.common.conflict

import groovy.transform.CompileStatic
import groovy.transform.NamedVariant
import io.tswf.transmutations.common.exception.ConflictResolveException
import org.objectweb.asm.tree.ClassNode
import org.objectweb.asm.tree.MethodNode

@CompileStatic
class ConflictResolver {

    @Lazy
    static Map<ConflictingMethodResolveStrategy, IConflictResolveStrategyLogic> conflictResolvers = ServiceLoader.load(IConflictResolveStrategyLogic)
        .toMapBy { it.enumValue.ensureNotNull { new NoSuchElementException("Can't find enum value for ${it}") } }
        .tap { resolvers ->
            def missingKeys = ConflictingMethodResolveStrategy.values()
                .stream()
                    .filter { !resolvers.containsKey(it) }
                    .toSet()

            if (missingKeys.isNotNullOrEmpty()) {
                throw new RuntimeException("There is no conflict resolve logic found for strategies: ${missingKeys}")
            }
        }

    @NamedVariant
    static boolean resolveConflict(
            ClassNode targetClass,
            MethodNode conflictingMethod,
            ClassNode anotherMethodOwner,
            MethodNode anotherMethod,
            IConflictingMethodResolveStrategyProvider strategyProvider
    ) {
        return resolveConflictUsingStratery(targetClass, conflictingMethod, anotherMethodOwner, anotherMethod, strategyProvider.conflictStrategy)
    }

    @NamedVariant
    static boolean resolveConflictUsingStratery(
            ClassNode targetClass,
            MethodNode conflictingMethod,
            ClassNode anotherMethodOwner,
            MethodNode anotherMethod,
            ConflictingMethodResolveStrategy strategy
    ) {
        try {
            def conflictResolver = conflictResolvers[strategy]
                    .ensureNotNull { new RuntimeException("There is no conflict resolve strategy logic found for: ${strategy.name()}!") }

            def conflictResolveContext = new ConflictResolveContext(
                    targetClass: targetClass,
                    conflictingMethod: conflictingMethod,
                    anotherMethodOwner: anotherMethodOwner,
                    anotherMethod: anotherMethod
            )

            conflictResolver.handleConflictResolve(conflictResolveContext)

            return conflictResolveContext.canContinue

        } catch (Throwable e) {
            // Если выброшенное исключение запланировано, то его обрабатывать кастомно не надо
            throw new ConflictResolveException(
                    "Exception occurred while trying to resolve conflict for ${conflictingMethod.javaFullName} in ${targetClass.javaName} and ${anotherMethod.javaFullName} located in ${anotherMethodOwner.javaName}!",
                    e
            )
        }
    }

}
